Module basic.

Inductive nat: Set:=
  | O: nat
  | S: nat -> nat.

Check nat.

Print nat.

Definition three: nat:= S (S (S O)).

Definition pred (n: nat): nat:=
  match n with
  | O => O
  | (S n') => n'
  end.

Fixpoint plus (n m: nat): nat:=
  match n with
  | O => m
  | (S n') => S (plus n' m)
  end.

Notation "a + b":= (plus a b) (at level 50, left associativity).

Theorem O_left_id: forall n, O + n = n.
Proof. reflexivity. Qed.

Theorem O_right_id: forall n, n + O = n.
Proof.
intros n.
induction n as [|n'].
reflexivity.
simpl. rewrite -> IHn'. reflexivity.
Qed.

Theorem plus_S_n: forall n m, S n + m = S (n + m).
Proof. reflexivity. Qed.

Theorem plus_n_S: forall n m, n + S m = S (n + m).
Proof.
intros n m.
induction n as [|n'].
reflexivity.
rewrite -> plus_S_n. rewrite -> IHn'. reflexivity.
Qed.

Ltac ind_try:=
  intros;
    match goal with
    |[ n:nat |- _ ] => induction n;repeat(simpl;try(subst;reflexivity))
    end.

Theorem O_right_id': forall n, n + O = n.
Proof.
ind_try.
rewrite -> IHn.
reflexivity.
Qed.
(**)
Theorem impl_trans: forall P Q R: Prop, (P->Q)->(Q->R)->P->R.
Proof (fun (P Q R: Prop) => (fun (H1:P->Q) (H2:Q->R) (p:P) => H2 (H1 p))).

Inductive bool: Set:=
  | false: bool
  | true: bool.

Inductive list: Set:=
  | nil: list
  | cons: nat -> list -> list.

Definition compose: (list->nat)->(nat->bool)->list->bool.
intros T1 T2 T3.
apply T2;apply T1;apply T3. Qed.

End basic.
(**)
Module lang.

Import Nat.

Inductive id: Type:=
  | Id: nat -> id.

Definition eq_id id1 id2:=
  match id1,id2 with
    | Id n1, Id n2 => eqb n1 n2
  end.

Definition total_map (A:Type):= id -> A.

Definition t_empty {A:Type} (v: A): total_map A:=
  (fun _ => v).

Definition t_update {A:Type} (m: total_map A)
                    (x: id) (v: A):=
  fun x' => if eq_id x x' then v else m x'.

Definition state:= total_map nat.

Definition empty_state: state:=
  t_empty O.

Inductive aexp: Type:=
  | ANum: nat -> aexp
  | AId: id -> aexp
  | APlus: aexp -> aexp -> aexp
  | AMinus: aexp -> aexp -> aexp
  | AMult: aexp -> aexp -> aexp.

Definition X: id:= Id 0.
Definition Y: id:= Id 1.
Definition Z: id:= Id 2.

Inductive bexp: Type:=
  | BTrue: bexp
  | BFalse: bexp
  | BEq: aexp -> aexp -> bexp
  | BLe: aexp -> aexp -> bexp
  | BNot: bexp -> bexp
  | BAnd: bexp -> bexp -> bexp.

Fixpoint aeval (st: state) (a: aexp): nat:=
  match a with
  | ANum n => n
  | AId x => st x
  | APlus a1 a2 => (aeval st a1) + (aeval st a2)
  | AMinus a1 a2  => (aeval st a1) - (aeval st a2)
  | AMult a1 a2 => (aeval st a1) * (aeval st a2)
  end.

Fixpoint beval (st: state) (b: bexp): bool:=
  match b with
  | BTrue       => true
  | BFalse      => false
  | BEq a1 a2   => eqb (aeval st a1) (aeval st a2)
  | BLe a1 a2   => leb (aeval st a1) (aeval st a2)
  | BNot b1     => negb (beval st b1)
  | BAnd b1 b2  => andb (beval st b1) (beval st b2)
  end.

Inductive com: Type:=
  | CSkip: com
  | CAss: id -> aexp -> com
  | CSeq: com -> com -> com
  | CIf: bexp -> com -> com -> com
  | CWhile: bexp -> com -> com.

Notation "'SKIP'":=
  CSkip.
Notation "x '::=' a":=
  (CAss x a) (at level 60).
Notation "c1 ;; c2":=
  (CSeq c1 c2) (at level 80, right associativity).
Notation "'WHILE' b 'DO' c 'END'":=
  (CWhile b c) (at level 80, right associativity).
Notation "'IFB' c1 'THEN' c2 'ELSE' c3 'FI'":=
  (CIf c1 c2 c3) (at level 80, right associativity).

Reserved Notation "c1 '/' st '\\' st'" (at level 40, st at level 39).

Inductive ceval : com -> state -> state -> Prop :=
  | E_Skip : forall st,
      SKIP / st \\ st
  | E_Ass  : forall st a1 n x,
      aeval st a1 = n ->
      (x ::= a1) / st \\ (t_update st x n)
  | E_Seq : forall c1 c2 st st' st'',
      c1 / st  \\ st' ->
      c2 / st' \\ st'' ->
      (c1 ;; c2) / st \\ st''
  | E_IfTrue : forall st st' b c1 c2,
      beval st b = true ->
      c1 / st \\ st' ->
      (IFB b THEN c1 ELSE c2 FI) / st \\ st'
  | E_IfFalse : forall st st' b c1 c2,
      beval st b = false ->
      c2 / st \\ st' ->
      (IFB b THEN c1 ELSE c2 FI) / st \\ st'
  | E_WhileEnd : forall b st c,
      beval st b = false ->
      (WHILE b DO c END) / st \\ st
  | E_WhileLoop : forall st st' st'' b c,
      beval st b = true ->
      c / st \\ st' ->
      (WHILE b DO c END) / st' \\ st'' ->
      (WHILE b DO c END) / st \\ st''
  where "c1 '/' st '\\' st'" := (ceval c1 st st').

Example ceval_example1:
    (X ::= ANum 2;;
     IFB BLe (AId X) (ANum 1)
       THEN Y ::= ANum 3
       ELSE Z ::= ANum 4
     FI)
   / empty_state
   \\ (t_update (t_update empty_state X 2) Z 4).
Proof.
apply E_Seq with (t_update empty_state X 2).
apply E_Ass. reflexivity.
apply E_IfFalse.
reflexivity.
apply E_Ass. reflexivity.
Qed.

Definition Assertion := state -> Prop.

Definition hoare_triple
           (P:Assertion) (c:com) (Q:Assertion) : Prop :=
  forall st st',
       c / st \\ st'  ->
       P st  ->
       Q st'.
Notation "{{ P }}  c  {{ Q }}" :=
  (hoare_triple P c Q) (at level 90, c at next level)
  : hoare_spec_scope.

Definition assert_implies (P Q : Assertion) : Prop :=
  forall st, P st -> Q st.

Notation "P ->> Q" :=
  (assert_implies P Q) (at level 80) : hoare_spec_scope.
Open Scope hoare_spec_scope.

Definition assn_sub X a P : Assertion :=
  fun (st : state) =>
    P (t_update st X (aeval st a)).

Notation "P [ X |-> a ]" := (assn_sub X a P) (at level 10).

Theorem hoare_asgn : forall Q X a,
  {{Q [X |-> a]}} (X ::= a) {{Q}}.
Proof.
  unfold hoare_triple.
  intros Q X a st st' HE HQ.
  inversion HE. subst.
  unfold assn_sub in HQ. assumption.  Qed.

Example assn_example :
  {{(fun st => st X = 1) [X |-> ANum 1]}}
  (X ::= (ANum 1))
  {{fun st => st X = 1}}.
Proof.
  apply hoare_asgn.  Qed.